/*
 *      udpserver.c
 *      
 *      Copyright 2009 - 2010 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>
#include <sys/utsname.h>
#include "gost.h"
#include "common/serialize.h"
#include "udpintctl.h"
#include "dbstorage.h"
#include "udpserver.h"
#include "mainwindow.h"

#define UDP_DATAGRAM_MAX_SIZE 1400

#define NETANS_SVCLIST_GET_XML "<netanswer ip=\"%s\" nodename=\"%s\">\n\t<request>%s</request>\n\t<result>%u</result>\n\t<datasize>%u</datasize>\n</netanswer>\n\n"
#define NETANS_SIGSTAT_GET_XML "<netanswer ip=\"%s\" nodename=\"%s\">\n\t<name>%s</name>\n\t<snr>%u</snr>\n\t<agc>%u</agc>\n\t<ber>%u</ber>\n\t<lock>%u</lock>\n\t<carrier>%u</carrier>\n</netanswer>\n\n"
#define NETANS_SVCTUNE_GET_XML "<netanswer ip=\"%s\" nodename=\"%s\">\n\t<request>%s</request>\n\t<result>%u</result>\n</netanswer>\n\n"
#define NETANS_GENERAL_ERROR   "<netanswer ip=\"%s\" nodename=\"%s\">\n\t<request>%s</request>\n\t<error>%s</error>\n</netanswer>\n\n"

struct _netgate *servernetgate;

enum {
	XML_PARAM_REQUEST = 0,
	XML_PARAM_FRAGMENT,
	XML_PARAM_SVCID,
	XML_PARAM_NULL
};

gint udpserver_xml_parameter = XML_PARAM_NULL;

enum {
	REQUEST_SIGSTAT_GET = 0,
	REQUEST_SVCLIST_GET,
	REQUEST_SVCLIST_MPT,
	REQUEST_SVCTUNE_GET,
	REQUEST_WRONG_PASSWORD,
	REQUEST_UNKNOWN
};

struct _request {
	//gchar password[64];
	gchar ip[32];
	guint request;
	guint fragment;
	guint svcid;
} request;


static void udpserver_xml_get_element(GMarkupParseContext *context, const gchar *element, const gchar **attr_names, const gchar **attr_values, gpointer user_data, GError **error);
static void udpserver_xml_get_value(GMarkupParseContext *context, const gchar *text, gsize size, gpointer data, GError **error);
static void udpserver_xml_end_element(GMarkupParseContext *context, const gchar *element, gpointer data, GError **error);


gint udpserver_run(struct _netgate *netgate)
{
	gint ready;
	struct sockaddr_in client;
	//select
	fd_set readset;
	struct timeval tv;
	//UDP integrity control
	static struct msghdr msgsend, msgrecv;
	struct iovec iovsend[2], iovrecv[2];
	//Serialization
	const gchar udpheader[] = UDP_HEADER_FORMAT;
	guint headersize;
	gchar *header;
	gchar *sndheader;
	gchar hdrflags;
	guint hdrseq;
	guint64 hdrtimestamp;
	guint hdrimito;
	guint64 hdrsyncro;
	//Buffers	
	gchar sendbuf[1400], recvbuf[1400], compbuf[1400], cryptbuf[1400];
	gsize sendbytes, recvbytes;
	gulong comprlen;
	gsize bytes;
	//XML Parser
	GMarkupParser mp;
	GMarkupParseContext *mpc;
	GError *error = NULL;
	//Other
	gchar *mptfragment;
	//signal status
	gchar name[256];
	guint16 snr;
	guint16 agc;
	guint32 ber;
	gboolean lock;
	gboolean carrier;
	
	headersize = serialize_get_size((char *)&udpheader);
	header = g_malloc0((gsize)headersize);
	
	while (1) {
		FD_ZERO(&readset);
		FD_SET(netgate->serversock, &readset);
		
		ready = select(netgate->serversock+1, &readset, NULL, NULL, NULL);
		
		if ((ready) && (FD_ISSET(netgate->serversock, &readset))) {
			msgrecv.msg_name = &client;
			msgrecv.msg_namelen = sizeof(struct sockaddr_in);
			msgrecv.msg_iov = iovrecv;
			msgrecv.msg_iovlen = 2;
			iovrecv[0].iov_base = header;
			iovrecv[0].iov_len = headersize;
			iovrecv[1].iov_base = &recvbuf;
			iovrecv[1].iov_len = sizeof(recvbuf);
			
			memset(header, 0, headersize);
			memset(recvbuf, 0, sizeof(recvbuf));
			
			bytes = recvmsg(netgate->serversock, &msgrecv, 0);
			
			serialize_parse_message((char *)&udpheader, header, TRUE, &hdrflags, &hdrseq, &hdrtimestamp, &hdrimito, &hdrsyncro);
			
			if (hdrflags == 1) {
				if (gost_segment_decrypt((gchar *)&recvbuf, (gchar *)&cryptbuf, bytes - headersize, (guint *)&netgate->serverhash, (guint *)&hdrimito)) {
					g_memmove(recvbuf, cryptbuf, bytes - headersize);
					
					comprlen = sizeof(compbuf);
					if (uncompress(compbuf, &comprlen, recvbuf, bytes - headersize) == Z_OK) {
						bytes = comprlen;
						g_memmove(recvbuf, compbuf, comprlen);
					}
					
					mp.start_element = udpserver_xml_get_element;
					mp.end_element = udpserver_xml_end_element;
					mp.text = udpserver_xml_get_value;
					mp.passthrough = NULL;
					mp.error = NULL; 
					
					mpc = g_markup_parse_context_new(&mp, 0, &request, NULL);
					g_markup_parse_context_parse(mpc, recvbuf, -1, &error);
					if (error != NULL) {
						g_warning("%s", error->message);
						g_error_free(error);
						g_markup_parse_context_free(mpc);
						request.request = REQUEST_UNKNOWN;
					} else {
						g_markup_parse_context_free(mpc);
					}	
				} else {
					request.request = REQUEST_WRONG_PASSWORD;
				}
			}
			
			memset(sendbuf, 0, sizeof(sendbuf));
			
			switch (request.request) {
				case REQUEST_SIGSTAT_GET:
					mainwindow_get_signal_status((gchar *)&name, &snr, &agc, &ber, &lock, &carrier);
					sendbytes = sprintf(sendbuf, NETANS_SIGSTAT_GET_XML, netgate->serveripstr, netgate->servernodename, name, snr, agc, ber, lock, carrier);
					break;
					
				case REQUEST_SVCLIST_GET:
					//Get permission for reading channels list and get its size
					if ((service_xml_list_get_length() > 0) && (netgate->serverexpsvclist)) {
						sendbytes = sprintf(sendbuf, NETANS_SVCLIST_GET_XML, netgate->serveripstr, netgate->servernodename, "SVCLIST_GET", 1, service_xml_list_get_length());
					} else {
						sendbytes = sprintf(sendbuf, NETANS_SVCLIST_GET_XML, netgate->serveripstr, netgate->servernodename, "SVCLIST_GET", 0, 0);
					}
					break;
					
				case REQUEST_SVCLIST_MPT:
					//Read fragment of channels list
					if ((service_xml_list_get_length() > 0) && (netgate->serverexpsvclist)) {
						if ((mptfragment = service_xml_list_get_fragment(UDP_DATAGRAM_MAX_SIZE, request.fragment)) == NULL) {
							sendbytes = sprintf(sendbuf, NETANS_GENERAL_ERROR, netgate->serveripstr, netgate->servernodename, "SVCLIST_MPT", "No such multipart fragment");
						} else {
							sendbytes = strlen(mptfragment);
							strncpy((gchar *)&sendbuf, mptfragment, sendbytes);
							g_free(mptfragment);
						}
					} else {
						sendbytes = sprintf(sendbuf, NETANS_GENERAL_ERROR, netgate->serveripstr, netgate->servernodename, "SVCLIST_MPT", "No permission to do it");
					}
					break;
					
				case REQUEST_SVCTUNE_GET:
					//Tune to channel
					if ((service_xml_list_get_length() > 0) && (netgate->serverexpsvclist)) {
						mainwindow_tune_start(request.svcid, SERVICE_TV_RADIO);
						sendbytes = sprintf(sendbuf, NETANS_SVCTUNE_GET_XML, netgate->serveripstr, netgate->servernodename, "SVCTUNE_GET", 1);
					} else {
						sendbytes = sprintf(sendbuf, NETANS_SVCTUNE_GET_XML, netgate->serveripstr, netgate->servernodename, "SVCTUNE_GET", 0);
					}
					break;
				
				case REQUEST_UNKNOWN:
					sendbytes = sprintf(sendbuf, NETANS_GENERAL_ERROR, netgate->serveripstr, netgate->servernodename, "UNKNOWN", "Unknown request");
					break;
					
				case REQUEST_WRONG_PASSWORD:
					sendbytes = sprintf(sendbuf, NETANS_GENERAL_ERROR, netgate->serveripstr, netgate->servernodename, "UNKNOWN", "Wrong password");
					break;
				
				default:
					sendbytes = sprintf(sendbuf, NETANS_GENERAL_ERROR, netgate->serveripstr, netgate->servernodename, "UNKNOWN", "Unknown request");
					break;
			}
			
			comprlen = sizeof(compbuf);
			if (compress(compbuf, &comprlen, sendbuf, sendbytes) == Z_OK) {
				sendbytes = comprlen;
				g_memmove(sendbuf, compbuf, sendbytes);
			}
			
			if (request.request != REQUEST_WRONG_PASSWORD) {
				if (gost_segment_encrypt((gchar *)&sendbuf, (gchar *)&cryptbuf, sendbytes, (guint *)&netgate->serverhash, (guint *)&hdrimito)) {
					g_memmove(sendbuf, cryptbuf, sendbytes);
					hdrflags = 1;
				}
			}
			
			sndheader = serialize_form_message((char *)&udpheader, TRUE, NULL, hdrflags, hdrseq, hdrtimestamp, hdrimito, hdrsyncro);
			
			msgsend.msg_name = &client;
			msgsend.msg_namelen = sizeof(struct sockaddr_in);
			msgsend.msg_iov = iovsend;
			msgsend.msg_iovlen = 2;
			iovsend[0].iov_base = sndheader;
			iovsend[0].iov_len = headersize;
			iovsend[1].iov_base = &sendbuf;
			iovsend[1].iov_len = sendbytes;
			
			bytes = sendmsg(netgate->serversock, &msgsend, 0);
			
			free(sndheader);
			
		} else continue;
		
	}
	
	g_free(header);
}

gint udpserver_start(struct _netgate *netgate)
{
	struct sockaddr_in server, self;
	struct hostent *lhp;
	struct utsname hostparams;
	gboolean started = FALSE;
	
	servernetgate = netgate;
	
	do {
		if ((netgate->serversock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
			perror("[-] Opening server socket");
			sleep(1);
			continue;
		}
		
		memset(&server, 0, sizeof(struct sockaddr_in));
		server.sin_family = AF_INET;
		server.sin_addr.s_addr = INADDR_ANY;
		server.sin_port = htons(5555);
		
		if (bind(netgate->serversock, (struct sockaddr *)&server, sizeof(struct sockaddr_in)) < 0) {
			perror("[-] Binding server socket");
			close(netgate->serversock);
			sleep(1);
			continue;
		}
		
		started = TRUE;
		
		gost_hash_new((gchar *)settings_read_string("local_server_password", "defaultpassword"), (guchar *)&netgate->serverhash);
		netgate->serverexpsvclist = settings_read_boolean("local_server_export_services", TRUE);
		
		
		if (uname(&hostparams) == 0) {
			if ((lhp = gethostbyname(hostparams.nodename)) != NULL) {
				memset(&self, 0, sizeof(struct sockaddr_in));
				self.sin_family = AF_INET;
				bcopy((char *)lhp->h_addr, (char *)&self.sin_addr, lhp->h_length);
				self.sin_port = htons(5555);
				
				inet_ntop(AF_INET, &(self.sin_addr), netgate->serveripstr, sizeof(netgate->serveripstr));
			} else {
				perror("[-] Server host unrecognized");
			}
			
			memset(netgate->servernodename, 0, sizeof(netgate->servernodename));
			strncpy(netgate->servernodename, hostparams.nodename, strlen(hostparams.nodename) > 256 ? 255 : strlen(hostparams.nodename));
			
		} else {
			perror("[-] Executing uname");
		}
	} while (!started);
	
	udpserver_run(netgate);
	
}

void udpserver_set_parameters(gchar *password, gboolean expsvclist)
{
	if ((password == NULL) || (strlen(password) == 0)) return;
	
	gost_hash_new(password, (guchar *)&servernetgate->serverhash);
	settings_append_string("local_server_password", password);
	servernetgate->serverexpsvclist = expsvclist;
	settings_append_boolean("local_server_export_services", expsvclist);
}

static void udpserver_xml_get_element(GMarkupParseContext *context, const gchar *element, const gchar **attr_names, const gchar **attr_values, gpointer data, GError **error)
{
	struct _request *request = data;
	
	memset(request->ip, 0, sizeof(request->ip));
	if ((attr_names[0] != NULL) && (strstr(attr_names[0], "ip") != NULL)) {
		if (attr_values[0] != NULL) {
			strncpy(request->ip, attr_values[0], strlen(attr_values[0]) > 64 ? 63 : strlen(attr_values[0]));
		}
	}
		
	if (g_str_equal(element, "request")) udpserver_xml_parameter = XML_PARAM_REQUEST;
	else if (g_str_equal(element, "fragment")) udpserver_xml_parameter = XML_PARAM_FRAGMENT;
	else if (g_str_equal(element, "svcid")) udpserver_xml_parameter = XML_PARAM_SVCID;
	else udpserver_xml_parameter = XML_PARAM_NULL;
}

static void udpserver_xml_get_value(GMarkupParseContext *context, const gchar *text, gsize size, gpointer data, GError **error)
{
	struct _request *request = data;
	
	if (udpserver_xml_parameter == XML_PARAM_NULL) return;
	
	switch (udpserver_xml_parameter) {
		case XML_PARAM_REQUEST:
			if (strstr(text, "SVCLIST_GET") != NULL) {
				request->request = REQUEST_SVCLIST_GET;
			} else if (strstr(text, "SVCLIST_MPT") != NULL) {
				request->request = REQUEST_SVCLIST_MPT;
			} else if (strstr(text, "SIGSTAT_GET") != NULL) {
				request->request = REQUEST_SIGSTAT_GET;
			} else if (strstr(text, "SVCTUNE_GET") != NULL) {
				request->request = REQUEST_SVCTUNE_GET;
			} else {
				request->request = REQUEST_UNKNOWN;
			}
			break;
		case XML_PARAM_FRAGMENT:
			request->fragment = strtoul(text, (gchar **)NULL, 10);
			break;
		case XML_PARAM_SVCID:
			request->svcid = strtoul(text, (gchar **)NULL, 10);
			break;
		default: break;
	}
	
}

static void udpserver_xml_end_element(GMarkupParseContext *context, const gchar *element, gpointer data, GError **error)
{
	udpserver_xml_parameter = XML_PARAM_NULL;
}
